from django.db import models
from django.core.validators import MinValueValidator, MaxValueValidator

from plugs_core.validators import NumberOfDigitsValidator


class NaiveNIFField(models.PositiveIntegerField):
    """
    Custom model field that represents a portuguese NIF
    this implementation is very naive, it only checks that the
    NIF contains 9 digits
    """

    default_validators = [NumberOfDigitsValidator(9)]

    def __init__(self, *args, **kwargs):
        super(NaiveNIFField, self).__init__(*args, **kwargs)


class PercentageField(models.FloatField):
    """
    Custom model field that represents a percentage in range 0-1
    in order to facilitate calculations with this value
    """

    default_validators = [MinValueValidator(0), MaxValueValidator(1)]

    def __init__(self, *args, **kwargs):
        super(PercentageField, self).__init__(*args, **kwargs)



"""
Temp Utils Module
"""

import re
import string
import random
import logging
from lxml import html

from django.conf import settings
from django.apps import apps
from django.core.exceptions import ObjectDoesNotExist

LOGGER = logging.getLogger(__name__)


def camel_to_snake(name):
    """
    Convert from CamelCase to snake_case
    http://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def get_db_distinct(queryset, field, func, **params):
    """
    Checks if a field / value pair exists in database
    and continues generating values until it finds
    a one that does not exist

    func is the function that generates values and
    params is the parameters that function takes
    """
    while True:
        try:
            value = func(**params)
            queryset.get(**{field: value})
        except ObjectDoesNotExist:
            break
    return value


def get_model_class(name):
    """
    This is being implemented to help
    with the Email Module, where we
    want to use a model for the email
    context without needing to import
    the model (which is most cases create
    a circular dependency, anyway)

    Beware that currently implementation
    returns the first match, so if a model
    with a same name exists in two different
    applications this will not work

    http://stackoverflow.com/a/13242421
    """
    LOGGER.warning('Beware, function returns first match in the model registry.')
    # iterate all registered models
    for model in apps.get_models():
        # return the app_label for first match
        if name == model._meta.object_name:
            app_label = model._meta.app_label
    return apps.get_model(app_label, name)


def get_non_field_errors_key():
    """
    Non field errors key can be override
    in the settings file, but is has a default value
    """
    try:
        key = settings.NON_FIELD_ERRORS_KEY
    except AttributeError:
        key = 'non_field_errors'
    return key


def html_to_text(html_string):
    """
    returns a plain text string when given a html string text
    handles a, p, h1 to h6 and br, inserts newline chars to
    create space in the string
    @todo handle images
    """
    # create a valid html document from string
    # beware that it inserts <hmtl> <body> and <p> tags
    # where needed
    html_tree = html.document_fromstring(html_string)

    # handle header tags
    for h in html_tree.cssselect("h1, h2, h3, h4, h5, h6"):
        # add two newlines after a header tag
        h.text = h.text + '\n\n'

    # handle links
    # find all a tags starting from the root of the document //
    # and replace the link with (link)
    for a in html_tree.xpath("//a"):
        href = a.attrib['href']
        a.text = a.text + " (" + href + ")"

    # handle paragraphs
    for p in html_tree.xpath("//p"):
        # keep the tail if there is one
        # or add two newlines after the text if there is no tail
        p.tail = p.tail if p.tail else "\n\n"

    # handle breaks
    for br in html_tree.xpath("//br"):
        # add a newline and then the tail (remaining text after the <br/> tag)
        # or add a newline only if there is no tail
        # http://stackoverflow.com/questions/18660382/how-can-i-preserve-br-as-newlines-with-lxml-html-text-content-or-equivalent?rq=1
        br.tail = "\n" + br.tail if br.tail else "\n"

    return html_tree.text_content()


def random_string(**kwargs):
    """
    By default generates a random string of 10 chars composed
    of digits and ascii lowercase letters. String length and pool can
    be override by using kwargs. Pool must be a list of strings
    """
    n = kwargs.get('length', 10)
    pool = kwargs.get('pool') or string.digits + string.ascii_lowercase
    return ''.join(random.SystemRandom().choice(pool) for _ in range(n))



from django.utils.translation import ugettext as _
from rest_framework import serializers

class NumberOfDigitsValidator(object):
    """
    Validates if the passed in integer has the
    right number of digits by converting the
    integer to a string and calling len() on it
    """

    message = _('Ensure number of digits is {0}.')

    def __init__(self, digits):
        self.digits = digits

    def __call__(self, value):
        number_of_digits = len(str(value))

        if number_of_digits != self.digits:
            message = self.message.format(self.digits)
            raise serializers.ValidationError(message)



"""
Custom Serializers Classes
"""

from rest_framework import serializers


class PrivateFieldsModelSerializer(serializers.ModelSerializer):
    """
    Subclasses ModelSerialzer to add the ability to restrict some
    fields when marked as private

    private_fields must be a tuple or list

    ...
    private_fields = ('email', )

    def get_serializer(self, *args, **kwargs):
        if self.request.method == 'GET':
            kwargs['private_fields'] = self.private_fields
        return super(UserViewSet, self).get_serializer(*args, **kwargs)
    """
    def __init__(self, *args, **kwargs):
        """
        Dealing with private fields
        """
        private = kwargs.pop('private_fields', None)
        super(PrivateFieldsModelSerializer, self).__init__(*args, **kwargs)
        if private is not None:
            for field_name in private:
                self.fields.pop(field_name)



from django.db import models
from django.utils.text import slugify
from django.core.exceptions import ObjectDoesNotExist

class Slugable(models.Model):

    slug = models.SlugField(max_length=75, null=False, unique=True)

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        """
        Overrides the save method
        """
        self.slug = self.create_slug()
        super(Slugable, self).save(*args, **kwargs)

    def create_slug(self):
        """
        Creates slug, checks if slug is unique, and loop if not
        """
        name = self.slug_source
        counter = 0
        # loops until slug is unique
        while True:
            if counter == 0:
                slug = slugify(name)
            else:
                # using the counter var to bump the slug name
                slug = slugify('{0} {1}'.format(name, str(counter)))
            try:
                # does the slug already exist, excluding the current object
                self.__class__.objects.exclude(pk=self.pk).get(slug=slug)
                # if slug exists increment counter and loop
                counter += 1
            except ObjectDoesNotExist:
                # the slug does not exist
                # we can break from the loop
                break
        return slug


class Timestampable(models.Model):
    """
    Adds created and updated fields
    """
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True



"""
Plugs API Test Case
"""

from rest_framework import status
from rest_framework.test import APITestCase
from rest_framework.response import Response

# TODO check this
# this may not be availabel
#try:
#    from post_office.models import Email
#except ImportError:
#    do something

from post_office.models import Email

from plugs_core.clients import PlugsAPIClient
from plugs_core import utils

def check_response_type(response):
    """
    Make sure a response object is an
    instance of REST Framework Response
    """
    assert isinstance(response, Response), 'response must be a rest framework response object'

# pylint: disable=C0103
# to keep the standard naming use in the rest of
# the assert methods we are creating invalid
# method names according to pylint
class PlugsAPITestCase(APITestCase):
    """
    Custom APITestCase that provides a APIClient with
    some nice methods and some extra assert methods
    for common use cases
    """

    client_class = PlugsAPIClient

    def assertFieldEqual(self, response, field, value):
        """
        Assert that the value of a field is equal to value
        """
        self.assertEqual(response.data.get(field), value)

    def assert200(self, response):
        """
        Assert responses status code is 200
        """
        check_response_type(response)
        self.assertEqual(status.HTTP_200_OK, response.status_code, response.data)

    def assert201(self, response):
        """
        Assert responses status code is 201
        """
        check_response_type(response)
        self.assertEqual(status.HTTP_201_CREATED, response.status_code, response.data)

    def assert204(self, response):
        """
        Assert responses status code is 204
        """
        check_response_type(response)
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code, response.data)

    def assert400(self, response):
        """
        Assert responses status code is 400
        """
        check_response_type(response)
        self.assertEqual(status.HTTP_400_BAD_REQUEST, response.status_code, response.data)

    def assert403(self, response):
        """
        Assert responses status code is 403
        """
        check_response_type(response)
        self.assertEqual(status.HTTP_403_FORBIDDEN, response.status_code, response.data)

    def assert404(self, response):
        """
        Assert responses status code is 404
        """
        check_response_type(response)
        self.assertEqual(status.HTTP_404_NOT_FOUND, response.status_code, response.data)

    def assert405(self, response):
        """
        Assert responses status code is 405
        """
        check_response_type(response)
        self.assertEqual(status.HTTP_405_METHOD_NOT_ALLOWED, response.status_code, response.data)

    def assertResultsLen(self, response, length):
        """
        Assert that the length of a list of results
        is equal to the lenght param

        len(response.data.get('results')) == length
        """
        check_response_type(response)
        assert isinstance(length, int)
        self.assertEqual(len(response.data.get('results')), length)

    def assertEmailCount(self, email_subject, email_count):
        """
        Assert that the number of emails in the database
        that match the criteria (for now just the subject) is
        equal to email_count
        """
        emails = Email.objects.filter(subject=email_subject)
        self.assertEqual(emails.count(), email_count)

    def assertEmailTo(self, email_subject, email_to, email_count=1):
        """
        Assert there is an email in the database with
        email_subject and email_to, we assume there is only
        one email matching these criteria
        """
        emails = Email.objects.filter(subject=email_subject)
        self.assertEqual(emails.count(), email_count)
        self.assertTrue(emails.get(to=email_to))

    def assertRequiredFields(self, response, required_fields, status_code=400):
        """
        Assert response has the status_code response status and the
        message is requiring the fields specified in the
        required_fields dict. By default status code is 400
        """
        check_response_type(response)
        for field in required_fields:
            self.assertContains(response, field, status_code=status_code)

    def assertNonFieldErrors(self, response, message, exact=True):
        """
        The non field errors is a list, it can have multiple
        messages, currenlty we are checking that it has only one
        brefore we proceed

        We check that the response status code must be 400

        The exact named argument is not currenly being used but the
        idea is to allow a list of errors to be checked, if exact
        is True, message should turn into a list of messages, and
        the number of messages must match the number of nfes
        """
        check_response_type(response)
        self.assert400(response)
        assert isinstance(message, str)
        nfe_key = utils.get_non_field_errors_key()
        nfe = response.data.get(nfe_key)
        assert len(nfe) == 1, 'We cannot deal with list with multiple elements.'
        self.assertEqual(nfe[0], message)

    def assertNonVisibleFields(self, response, fields, status_code=200):
        """
        Assert that fields are not visible in the response.
        By default check that response has status code 200
        """
        check_response_type(response)
        for field in fields:
            self.assertNotContains(response, field, status_code=status_code)

    def assertLogin(self, username, password):
        """
        Assert login, aka, exchange username and password for jwt
        """
        data = {'email': username, 'password': password}
        response = self.client.post('/api-token-auth/', data)
        self.assert200(response)  # to provide debug info
        self.assertContains(response, 'token', status_code=200)

    def assertCannotLogin(self, username, password):
        """
        Assert cannot login, aka, cannot exchange username and password for jwt
        """
        data = {'email': username, 'password': password}
        response = self.client.post('/api-token-auth/', data)
        self.assert400(response)  # to provide debug info

    def assertObjFieldEqual(self, obj, field, value):
        """
        Assert field in obj is equal to provided value
        """
        self.assertEqual(getattr(obj, field), value)



"""
Core Permissions Classes
"""

from rest_framework import permissions

class SelfOrReadOnly(permissions.BasePermission):
    """
    Object-level permission to only allow owners of an object to edit it.
    ATTENTION: This only works for user objects, future implementations should extend this
    """

    def has_object_permission(self, request, view, obj):
        # Read permissions are allowed to any request,
        # so we'll always allow GET, HEAD or OPTIONS requests.
        if request.method in permissions.SAFE_METHODS:
            return True

        return obj == request.user


class IsOwnerOrReadOnly(permissions.BasePermission):
    """
    Object-level permission to only allow owners of an object to edit it.
    This only works for objects related to a user through a field user
    """

    def has_object_permission(self, request, view, obj):
        # Read permissions are allowed to any request,
        # so we'll always allow GET, HEAD or OPTIONS requests.
        if request.method in permissions.SAFE_METHODS:
            return True

        return obj.user == request.user



"""
Custom API Client
"""

from urllib.parse import urlencode
from rest_framework.test import APIClient

class PlugsAPIClient(APIClient):
    """
    Subclasses APIClient to provide
    custom functionality
    """

    def parse_params(self, params):
        """
        Parsing params, params is a dict
        and the dict value can be a string
        or an iterable, namely a list, we
        need to process those iterables
        """
        for (key, value) in params.items():
            if not isinstance(value, str):
                string_params = self.to_string(value)
                params[key] = string_params
        return params

    def to_string(self, obj):
        """
        Picks up an object and transforms it
        into a string, by coercing each element
        in an iterable to a string and then joining
        them, or by trying to coerce the object directly
        """
        try:
            converted = [str(element) for element in obj]
            string = ','.join(converted)
        except TypeError:
            # for now this is ok for booleans
            string = str(obj)
        return string

    def filter(self, endpoint, params):
        """
        Makes a get request by construction
        the path from an endpoint and a dict
        with filter query params

        e.g.
        params = {'category__in': [1,2]}
        response = self.client.filter('/experiences/', params)
        """
        params = self.parse_params(params)
        params = urlencode(params)
        path = '{0}?{1}'.format(endpoint, params)
        return self.get(path)



"""
Core Viewsets
"""

from rest_framework.viewsets import GenericViewSet
from rest_framework.mixins import (
    CreateModelMixin, ListModelMixin, RetrieveModelMixin,
    DestroyModelMixin, UpdateModelMixin)


class CreateReadViewSet(CreateModelMixin, ListModelMixin, RetrieveModelMixin, GenericViewSet):
    """
    A viewset that provides create and read (list and retrieve) actions
    """
    pass

class CreateReadDestroyViewSet(CreateModelMixin, ListModelMixin, RetrieveModelMixin, DestroyModelMixin, GenericViewSet):
    """
    A viewset that provides create, read (list and retrieve) and destroy actions
    """
    pass

class UpdateReadViewSet(UpdateModelMixin, ListModelMixin, RetrieveModelMixin, GenericViewSet):
    """
    A viewset that provides update and read (list and retrieve) actions
    """
    pass

class CreateUpdateReadViewSet(CreateModelMixin, UpdateModelMixin, ListModelMixin, RetrieveModelMixin, GenericViewSet):
    """
    A viewset that provides create, update and read (list and retrieve) actions
    """
    pass

class CreateViewSet(CreateModelMixin, GenericViewSet):
    """
    A viewset that provides create action
    """
    pass
